import json
from typing import Any, Dict, Optional

from _pytest.monkeypatch import MonkeyPatch
from pytest import fixture, mark, raises

from melissa import MELISSA_URL, ApiException
from melissa import AsyncMelissa as Melissa
from tests.helpers import load_fixture

AUTH_LOGIN_URL = MELISSA_URL % "auth/login"
STATUS_URL = MELISSA_URL % "provider/fetch"
FETCH_DEVICES_URL = MELISSA_URL % "controllers"
FETCH_GEOFENCES_URL = MELISSA_URL % "geofences"
SEND_URL = MELISSA_URL % "provider/send"
CUR_SETTINGS_URL = MELISSA_URL % "controllers/12345678"

LOGGED_IN_HEADERS = {
    "Accept": "application/json, text/plain, */*",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,sv-SE;q=0.9",
    "Authorization": "Bearer 12345678",
    "Cache-Control": "no-cache",
    "Connection": "keep-alive",
    "Pragma": "no-cache",
    "User-Agent": "Mozilla/5.0 (Linux; Android 7.0; SM-G935F Build/NRD90M; wv)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 "
    "Chrome/63.0.3239.111 Mobile Safari/537.36",
    "X-Requested-With": "at.cloudfaces.melissa",
}

FETCH_DEVICES_DATA_OK: Dict[str, Dict[str, Any]] = {
    "12345678": {
        "_links": {"self": {"href": "/v1/controllers"}},
        "brand_id": 1,
        "controller_log": {
            "created": "2018-01-16T18:02:36.735Z",
            "humidity": 15.3,
            "raw_humidity": 11142,
            "raw_temperature": 29384,
            "temp": 28.6,
        },
        "created": "2016-07-06 18:59:46",
        "firmware_version": "V1SHTHF",
        "id": 1,
        "mac": "12345678",
        "name": "Melissa 12345678",
        "online": True,
        "room_id": None,
        "serial_number": "12345678",
        "type": "melissa",
        "user_id": 1,
    }
}

FETCH_GEOFENCES_DATA_OK: Dict[str, Dict[str, Any]] = {
    "12345678": {
        "_links": {"self": {"href": "/v1/geofences/1"}},
        "active": 1,
        "controller_id": "12345678",
        "enter_command_fields": '{"state":1,"mode":"2","temp":30,"fan":"3"}',
        "enter_command_name": "send_ir_code",
        "enter_push": 1,
        "exit_command_fields": '{"state":0,"mode":3,"temp":25,"fan":0}',
        "exit_command_name": "send_ir_code",
        "exit_push": 1,
        "id": 1,
        "latitude": 1,
        "longitude": 1,
        "radius": 1500,
        "user_id": 1,
    }
}

CUR_SETTINGS_DATA_OK = {
    "_links": {"self": {"href": "/v1/controllers/12345678"}},
    "controller": {
        "_relation": {
            "command_log": {"fan": 2, "mode": 2, "state": 1, "temp": 16}
        },
        "created": "2016-07-06 18:59:46",
        "deleted_at": None,
        "firmware_version": "V1SHTHF",
        "id": 1,
        "mac": "12345678",
        "name": "Melissa 12345678",
        "online": False,
        "room_id": None,
        "serial_number": "12345678",
        "type": "melissa",
        "user_id": 1,
    },
}


class MockResponse:
    def __init__(self, text: Optional[str], status_code: int) -> None:
        self._text = text
        self.status = status_code

    async def text(self) -> Optional[str]:
        return self._text

    def result(self) -> "MockResponse":
        return self


class MockedClientSession:
    def __init__(*args: Any, **kwargs: Any) -> None:
        """Dummy init method"""

    async def post(self, *args: Any, **kwargs: Any) -> MockResponse:
        if args[0] == AUTH_LOGIN_URL:
            return MockResponse(load_fixture("auth_login.json"), 200)
        elif args[0] == STATUS_URL:
            return MockResponse(load_fixture("status.json"), 200)
        elif args[0] == SEND_URL:
            return MockResponse(load_fixture("send.json"), 200)

        return MockResponse(None, 404)

    async def get(self, *args: Any, **kwargs: Any) -> MockResponse:
        print(args)
        if args[0] == FETCH_DEVICES_URL:
            return MockResponse(load_fixture("fetch_devices.json"), 200)
        elif args[0] == FETCH_GEOFENCES_URL:
            return MockResponse(load_fixture("fetch_geofences.json"), 200)
        elif args[0] == CUR_SETTINGS_URL:
            return MockResponse(load_fixture("cur_settings.json"), 200)

        return MockResponse(None, 404)


@fixture(autouse=True)
def mocked_aiohttp(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr("melissa.ClientSession", MockedClientSession)


async def mocked_post_bad(*args: Any, **kwargs: Any) -> MockResponse:
    if args[0] == AUTH_LOGIN_URL:
        return MockResponse(load_fixture("auth_login_denied.json"), 401)
    elif args[0] == SEND_URL:
        return MockResponse(load_fixture("error.json"), 500)
    elif args[0] == STATUS_URL:
        return MockResponse(load_fixture("bad_temp_status.json"), 200)

    return MockResponse(None, 404)


def test_init() -> None:
    melissa = Melissa(
        username="1234",
        password="4321",
        access_token="12345678",
        refresh_token="12345678",
    )
    assert melissa.username == "1234"
    assert melissa.password == "4321"
    assert melissa.access_token == "12345678"
    assert melissa.refresh_token == "12345678"


@mark.asyncio
async def test_connect_bad(monkeypatch: MonkeyPatch) -> None:
    melissa = Melissa(username="1234", password="4321")
    monkeypatch.setattr("melissa.ClientSession.post", mocked_post_bad)
    with raises(ApiException):
        await melissa.async_connect()


def test_get_headers() -> None:
    melissa = Melissa(
        username="1234",
        password="4321",
        access_token="12345678",
        token_type="Bearer",
    )
    assert melissa._get_headers() == LOGGED_IN_HEADERS


def test_sanity_check() -> None:
    melissa = Melissa(username="1234", password="4321")
    melissa._latest_temp = 28.9
    melissa._latest_status["12345678"] = json.loads(
        load_fixture("status.json")
    )["provider"]
    data = json.loads(load_fixture("bad_temp_status.json"))["provider"]
    device = "12345678"
    assert not melissa.sanity_check(data, device)
    data = json.loads(load_fixture("bad_hum_status.json"))["provider"]
    assert not melissa.sanity_check(data, device)


@mark.asyncio
async def test_fetch_devices() -> None:
    melissa = Melissa(username="1234", password="4321")
    resp = await melissa.async_fetch_devices()
    assert resp == FETCH_DEVICES_DATA_OK


@mark.asyncio
async def test_fetch_geofences() -> None:
    melissa = Melissa(username="1234", password="4321")
    resp = await melissa.async_fetch_geofences()
    assert resp == FETCH_GEOFENCES_DATA_OK


def test_have_connection() -> None:
    melissa = Melissa(
        username="1234", password="4321", access_token="12345678"
    )
    assert melissa.have_connection
    melissa.access_token = None
    assert not melissa.have_connection


@mark.asyncio
async def test_send_ok() -> None:
    melissa = Melissa(username="1234", password="4321")
    assert await melissa.async_send("12345678", "melissa", {"temp": 20})


@mark.asyncio
async def test_send() -> None:
    melissa = Melissa(username="1234", password="4321")
    assert await melissa.async_send("12345678", "melissa", {"temp": 20})


@mark.asyncio
async def test_status() -> None:
    melissa = Melissa(username="1234", password="4321")
    good_status = await melissa.async_status()
    melissa.fetch_timestamp = None
    bad_status = await melissa.async_status()
    assert good_status == bad_status


@mark.asyncio
async def test_cur_settings() -> None:
    melissa = Melissa(username="1234", password="4321")
    assert await melissa.async_cur_settings("12345678") == CUR_SETTINGS_DATA_OK



"""Python implementation of the mclimate api."""

import json
import logging
from datetime import UTC, datetime
from importlib.metadata import PackageNotFoundError, version
from numbers import Number
from typing import Any, Dict, Optional, Union

from aiohttp import ClientSession

from .const import (
    CHANGE_THRESHOLD,
    CHANGE_TIME_CACHE_DEFAULT,
    CLIENT_DATA,
    HEADERS,
    MELISSA_URL,
    MIN_HUMIDITY_ALLOWED,
)
from .exceptions import ApiException, UnsupportedDevice

try:
    __version__ = version("melissa")
except PackageNotFoundError:
    __version__ = "dev"

__author__ = "Magnus Knutas"

_LOGGER = logging.getLogger(__name__)

__all__ = ("ApiException", "AsyncMelissa", "MELISSA_URL")


class AsyncMelissa:
    """Async class for Melissa."""

    SERIAL_NUMBER: str = "serial_number"
    COMMAND: str = "command"
    STATE: str = "state"
    MODE: str = "mode"
    TEMP: str = "temp"
    FAN: str = "fan"
    HUMIDITY: str = "humidity"

    FAN_AUTO: int = 0
    FAN_LOW: int = 1
    FAN_MEDIUM: int = 2
    FAN_HIGH: int = 3

    STATE_OFF: int = 0
    STATE_ON: int = 1
    STATE_IDLE: int = 2

    MODE_AUTO: int = 0
    MODE_FAN: int = 1
    MODE_HEAT: int = 2
    MODE_COOL: int = 3
    MODE_DRY: int = 4

    LED_ON: str = "reset_color"
    LED_OFF: str = "turn_led_off"

    DEFAULT_DATA_MELISSA: Dict[str, Union[int, str]] = {
        SERIAL_NUMBER: "",
        COMMAND: "send_ir_code",
        STATE: STATE_ON,  # 0-OFF; 1-ON; 2-Idle
        MODE: MODE_HEAT,  # 0-Auto; 1-Fan; 2-Heat; 3-Cool; 4-Dry
        TEMP: 20,  # number between 16 and 30 (depends on the codeset)
        FAN: FAN_MEDIUM,  # 0-Auto; 1-Low; 2-Med; 3-High
    }

    DEFAULT_DATA_BOBBIE: Dict[str, str] = {
        SERIAL_NUMBER: "",
        COMMAND: "switch_on_off",
        STATE: "on",
    }

    DEFAULT_DATA_LED: Dict[str, str] = {SERIAL_NUMBER: "", COMMAND: LED_ON}

    def __init__(
        self,
        username: str,
        password: str,
        headers: Dict[str, str] = HEADERS,
        time_cache: int = CHANGE_TIME_CACHE_DEFAULT,
        access_token: Optional[str] = None,
        token_type: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ) -> None:
        """Init method."""
        self.default_headers: Dict[str, str] = headers
        self.access_token: Optional[str] = access_token
        self.token_type: Optional[str] = token_type
        self._latest_status: Dict[str, Any] = {}
        self.session: ClientSession = ClientSession(connector_owner=False)
        self.username: str = username
        self.password: str = password
        self.refresh_token: Optional[str] = refresh_token
        self.devices: Dict[str, Dict[str, Any]] = {}
        self.geofences: Dict[str, Dict[str, Any]] = {}
        self._latest_humidity: Optional[float] = None
        self._latest_temp: Optional[float] = None
        self._time_cache: int = time_cache
        self.fetch_timestamp: Optional[datetime] = None

    def _get_headers(self) -> Dict[str, str]:
        headers = self.default_headers.copy()
        headers.update(
            {"Authorization": "%s %s" % (self.token_type, self.access_token)}
        )
        _LOGGER.debug(headers)
        return headers

    def sanity_check(self, data: Dict[str, Any], device: str) -> bool:
        """Sanity check method."""
        ret = True
        _latest: Optional[Dict[str, Union[str, Number, None]]] = (
            self._latest_status.get(device)
        )
        if _latest is not None:
            _LOGGER.debug(data)
            left = data[self.TEMP]
            right = _latest[self.TEMP]
            if abs(left - right) > CHANGE_THRESHOLD:
                ret = False
        if (
            self._latest_status
            and self._latest_status.get(device)
            and data[self.HUMIDITY] < MIN_HUMIDITY_ALLOWED
        ):
            ret = False
        return ret

    @property
    def have_connection(self) -> bool:
        """Have connection method."""
        return self.access_token is not None

    async def async_connect(self) -> None:
        """Async connect method."""
        url = MELISSA_URL % "auth/login"
        _LOGGER.info(url)
        data = CLIENT_DATA.copy()
        data.update({"username": self.username, "password": self.password})
        _LOGGER.info(data)
        req = await self.session.post(url, data=data, headers=HEADERS)
        if req.status == 200:
            resp = json.loads(await req.text())
            self.access_token = resp["auth"]["access_token"]
            self.refresh_token = resp["auth"]["refresh_token"]
            self.token_type = resp["auth"]["token_type"]
        else:
            raise ApiException(await req.text(), req.status)

    async def async_fetch(self, url: str) -> str:
        """Async fetch method."""
        response = await self.session.get(url)
        ret: str = await response.text()
        return ret

    async def async_fetch_devices(self) -> Dict[str, Dict[str, Any]]:
        """Async fetch devices method."""
        url = MELISSA_URL % "controllers"
        _LOGGER.info(url)
        headers = self._get_headers()
        req = await self.session.get(url, headers=headers)
        if req.status == 200:
            resp = json.loads(await req.text())
            for controller in resp["_embedded"]["controller"]:
                self.devices[controller["serial_number"]] = controller
        _LOGGER.debug(self.devices)
        return self.devices

    async def async_fetch_geofences(self) -> Dict[str, Dict[str, Any]]:
        """Async fetch geofences method."""
        url = MELISSA_URL % "geofences"
        _LOGGER.info(url)
        headers = self._get_headers()
        req = await self.session.get(url, headers=headers)
        if req.status == 200:
            resp = json.loads(await req.text())
            for geofence in resp["_embedded"]["geofence"]:
                self.geofences[geofence["controller_id"]] = geofence
        _LOGGER.info(self.geofences)
        return self.geofences

    async def async_send(
        self,
        device: str,
        device_type: str = "melissa",
        state_data: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Async send method."""
        data: Dict[str, Any]
        if device_type == "melissa":
            data = self.DEFAULT_DATA_MELISSA.copy()
        elif device_type == "bobbie":
            data = self.DEFAULT_DATA_BOBBIE.copy()
        elif device_type == "led":
            data = self.DEFAULT_DATA_LED.copy()
        else:
            raise UnsupportedDevice(device_type)

        if state_data:
            data.update(state_data)

        data.update({self.SERIAL_NUMBER: device})
        url = MELISSA_URL % "provider/send"
        _LOGGER.info(url)

        if not self.have_connection:
            await self.async_connect()

        headers = self._get_headers()
        headers.update({"Content-Type": "application/json"})

        input_data = json.dumps(data)
        _LOGGER.debug(input_data)
        req = await self.session.post(url, data=input_data, headers=headers)
        if not req.status == 200:
            raise ApiException(await req.text(), req.status)

        return True

    async def async_status(
        self, test: bool = False, cached: bool = False
    ) -> Dict[str, Any]:
        """Async status method."""
        if (
            cached
            and self.fetch_timestamp
            and self._time_cache
            > (datetime.now(UTC) - self.fetch_timestamp).total_seconds()
        ):
            return self._latest_status
        url = MELISSA_URL % "provider/fetch"
        _LOGGER.info(url)
        headers = self._get_headers()
        headers.update({"Content-Type": "application/json"})
        ret = {}
        if not self.devices:
            await self.async_fetch_devices()
        for device in self.devices:
            if str(self.devices[device]["type"]) in ("melissa", "bobbie"):
                input_data = json.dumps({"serial_number": device})
                req = await self.session.post(
                    url, data=input_data, headers=headers
                )
                if req.status == 200:
                    data = json.loads(await req.text())
                    ret[device] = data["provider"]
                elif req.status == 401 and not test:
                    await self.async_connect()
                    return await self.async_status()
                else:
                    raise ApiException(await req.text())
        self.fetch_timestamp = datetime.now(UTC)
        self._latest_status = ret
        return ret

    async def async_cur_settings(self, serial_number: str) -> Dict[str, Any]:
        """Async cur settings method."""
        url = MELISSA_URL % "controllers/%s" % serial_number
        _LOGGER.info(url)
        headers = self._get_headers()
        req = await self.session.get(url, headers=headers)
        if req.status == 200:
            data: Dict[str, Any] = json.loads(await req.text())
        else:
            raise ApiException(await req.text(), req.status)
        _LOGGER.debug(data)
        return data



"""Exceptions for py-melissa-climate."""

from typing import Optional


class _Exception(Exception):
    def __init__(
        self, _message: str, _status_code: Optional[int] = None
    ) -> None:
        self._message = _message
        self._status_code = _status_code

    @property
    def message(self) -> Optional[str]:
        return self._message

    @property
    def status_code(self) -> Optional[int]:
        return self._status_code


class ApiException(_Exception):
    """Api error occured."""


class UnsupportedDevice(_Exception):
    """Unsupported device."""



"""Constants for Melissa platform."""

from typing import Dict

MELISSA_URL: str = "http://developer-api.seemelissa.com/v1/%s"
CLIENT_DATA: Dict[str, str] = {
    "client_id": "mclimate",
    "client_secret": "mclimate_core",
}
HEADERS: Dict[str, str] = {
    "User-Agent": "Mozilla/5.0 (Linux; Android 7.0; SM-G935F Build/NRD90M; wv)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0"
    " Chrome/63.0.3239.111 Mobile Safari/537.36",
    "Connection": "keep-alive",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
    "Accept": "application/json, text/plain, */*",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,sv-SE;q=0.9",
    "X-Requested-With": "at.cloudfaces.melissa",
}

CHANGE_THRESHOLD: int = 10
MIN_HUMIDITY_ALLOWED: int = 10
CHANGE_TIME_CACHE_DEFAULT: int = 120  # 2 min default
